"""
Eleos: A utility library for working with planetary spectra and gneerating NEMESIS radiative transfer cores.

Modules available directly under `eleos`:
- constants
- cores
- parsers
- profiles
- results
- shapes
- spectra
- spx
- utils
"""

# Explicitly declare the public API
__all__ = [
    "constants",
    "cores",
    "parsers",
    "profiles",
    "results",
    "shapes",
    "spectra",
    "spx",
    "utils",
]

# Import submodules into package namespace
from . import (
    constants,
    cores,
    parsers,
    profiles,
    results,
    shapes,
    spectra,
    spx,
    utils,
)


"""This module provides parsing objects for reading some NEMESIS files, such as nemesis.ref"""

import pandas as pd
import itertools as it
import io
import numpy as np
from pathlib import Path
import astropy.units as u
import struct

import warnings
warnings.formatwarning = lambda msg, *_: f"Warning: {msg}\n"

from . import utils
from . import constants
from . import profiles as profiles_
from . import shapes


## TODO: Move all parsing routines here, .itr, .prc etc...
## I also want to be able to completely reconstruct the core object from these files so no pickling of NemesisCore needed, but thats a little way off


class Parser:
    def __init__(self, filepath):
        self.filepath = Path(filepath)
        self.read()


class NemesisRef(Parser):
    """Parser for nemesis.ref and nemesis.prf
    
    Attributes:
        amform:    Number of latitudes (always 1)
        planet_id: ID of the planet being analysed.
        latitude:  Latitude at which the .ref file applies
        num_layers:Number of layers used 
        num_gases: Number of gases in the file
        gas_names: List of the names of all the gases
        data:      Contains the pressure, temperature, and VMR profiles of each gas at each height
    """
    def __init__(self, filepath):
        self._extra_header = True
        super().__init__(filepath)

    def read(self):
        with open(self.filepath) as file:
            lines = file.read().split("\n")
            if self._extra_header:
                del lines[1]
                
        self.amform, = utils.get_ints_from_string(lines[0])
        planet_id, latitude, num_layers, num_gases = utils.get_floats_from_string(lines[1])
        self.planet_id = int(planet_id)
        self.latitude = latitude
        self.num_layers = int(num_layers)
        self.num_gases = int(num_gases)

        self.gas_names = []
        for l in lines[2:2+int(self.num_gases)]:
            gas_id, isotope_id = utils.get_ints_from_string(l)
            gas_name = constants.GASES[constants.GASES.radtrans_id == gas_id].name.iloc[0]
            self.gas_names.append(f"{gas_name} {isotope_id}")

        self.data = pd.read_table(self.filepath, 
                                  skiprows=3+int(self._extra_header)+self.num_gases, 
                                  sep="\s+", 
                                  header=None)
        self.data.columns = ["height", "pressure", "temperature"] + self.gas_names


class NemesisPrf(NemesisRef):
    def __init__(self, filepath):
        self.filepath = Path(filepath)
        self._extra_header = False
        self.read()


class NemesisMre(Parser):
    """Parser for the nemesis.mre file
    
    Attributes:
        ispec (int): Don't know
        ngeom (int): Number of geometries (should be 1)
        latitude (float): Latitude of the observation
        longitude (float): Longitude of the observation
        retrieved_spectrum pd.DataFrame: DataFrame containing the measured spectrum + all error sources and the fitted model spectra and its errors
        retrieved_parameters List[pd.DataFrame]: List of DataFrames containing the retrieved parameters from each Profile"""
    
    def _parse_header_line(self, line, num_fields, cast_to):
        fields = [cast_to(x) for x in line.split()[:num_fields]]
        if num_fields == 1:
            return fields[0]
        else:
            return fields

    def read(self):
        with open(self.filepath) as file:
            mre_data = file.read().split("\n")

        header = []
        blocks = []
        for i, line in enumerate(mre_data):
            # Read the first 3 lines to the header array
            if i < 3:
                header.append(line)
            # Find the boundaries between result blocks
            elif "Variable" in line:
                blocks.append(i)
        blocks.append(i)

        # Set some attributes from the header info    
        self.ispec, self.ngeom, _,_,_ = self._parse_header_line(header[1], num_fields=5, cast_to=int)
        self.latitude, self.longitude = self._parse_header_line(header[2], num_fields=2, cast_to=float)

        # Read in the fitted spectrum as a DataFrame
        self.retrieved_spectrum = pd.read_table(self.filepath, 
                                                names=["wavelength", "measured", "error", "pct_error", "model", "pct_diff"],
                                                index_col=0, sep="\s+", skiprows=5, nrows=blocks[0]-7)

        # Read in each retrieved parameter 
        self.retrieved_parameters = []
        self.initial_state_vector = []
        with open(self.filepath) as file:
            for start, end in it.pairwise(blocks):
                data = utils.read_between_lines(file, start, end)
                df = pd.read_table(io.StringIO(data), skiprows=4, sep="\s+", names=["i", "ix", "prior", "prior_error", "retrieved", "retrieved_error"])
                df.drop(["i", "ix"], axis=1, inplace=True)
                self.initial_state_vector += list(df.prior)
                self.retrieved_parameters.append(df)
        

class NemesisXsc(Parser):
    """Parser for the nemesis.xsc file
    
    Attributes:
        xsc (pd.DataFrame): The aerosol cross-sections as a function of wavelength for each aerosol mode
        ssa (pd.DataFrame): The single scattering albedos as a function of wavelength for each aerosol modes"""
    
    def read(self):
        waves = []
        ssas = []
        xscs = []
        with open(self.filepath) as file:
            for i, line in enumerate(file):
                if i == 0:
                    continue
                if i % 2 == 1:
                    wavelength, *x = utils.get_floats_from_string(line)
                    waves.append(wavelength)
                    xscs.append(x)
                else:
                    s = utils.get_floats_from_string(line)
                    ssas.append(s)

        self.ssa = pd.DataFrame(ssas)
        self.ssa.insert(0, column="wavelength", value=waves)
        self.xsc = pd.DataFrame(xscs)
        self.xsc.insert(0, column="wavelength", value=waves)
            

class NemesisItr(Parser):
    """Parser for nemesis.itr. Also requires a NemesisMre parser
    
    Attributes:
        state_vectors (pd.DataFrame): Linear state vectors for each iteration"""
    
    def __init__(self, filepath, mre=None):
        if mre is None:
            try:
                fp = Path(filepath).parent / "nemesis.mre"
                self.mre = NemesisMre(fp)
            except:
                print("Warning: Failed to load {fp}, skipping state vector linearisation")
                self.mre = None
        else:
            self.mre = mre
        super().__init__(filepath)

    def read(self):
        data = []
        count = -1
        with open(self.filepath) as file:
            for i, line in enumerate(file.read().split("\n")):
                if line == " ":
                    count = 3
                else:
                    count -= 1
                if count == 0:
                    d = []
                    for i, v in enumerate(line.split()):
                        d.append(float(v))
                    data.append(d)

        if self.mre is not None:
            exps = []
            for i, value in enumerate(data[0]):
                flag = np.isclose(self.mre.initial_state_vector[i], np.exp(value))
                exps.append(flag)
        else:
            exps = [False] * len(data[0])

        self.state_vectors = pd.DataFrame(data)
        for i, column in enumerate(self.state_vectors.columns):
            if exps[i]:
                self.state_vectors[column] = np.exp(self.state_vectors[column])

    def add_column_names(self, profiles):
        names = []
        for label, profile in profiles.items():
            print(profile.VARIABLES)
            print(profile.shape.VARIABLES)
            if isinstance(profile.shape, shapes.Shape0):
                names += [f"{label} {n}" for n in range(120)] # temp value FIX
            else:
                names += [f"{label} {n}" for n in profile.VARIABLES]
                print()

        print("Done:")

        self.state_vectors.columns = names


class NemesisPrc(Parser):
    """Parser for nemesis.prc
    
    Attributes:
        chisq (List[float]): List of chi-squared values"""
    def read(self):
        self.chisq = []
        with open(self.filepath) as file:
            for line in file:
                if "chi" in line and "should" not in line:
                    self.chisq.append(utils.get_floats_from_string(line)[0])


class NemesisSpx(Parser):
    """Parser for the nemesis.spx file

    Attributes:
        lat (float):              Latitude of the observation
        lon (float):              longitude of the observation
        phase (float):            Phase angle of the observation
        emission (float):         Emission angle of the observation
        azimuth (float):          Azimuth angle of the observation
        wgeom (float):            wgeom
        nconv (int):              nconv
        nav (int):                nav
        fwhm (float):             FWHM of the instrument
        wavelengths (np.ndarray): Wavelengths of the observed spectra in um
        spectrum (np.ndarray):    The observed spectra in W/cm2/sr/um
        error (np.ndarray):       The error on the observed spectra in W/cm2/sr/um
    """

    def read(self):
        """Read the spx file in
        
        Args:
            None
            
        Returns:
            None
        """
        with open(self.filepath) as f:
            self.fwhm, *_ = (float(x) for x in f.readline().split())
            self.nconv = int(float(f.readline()))
            self.nav = int(float(f.readline()))
            self.lat, self.lon, self.phase, self.emission, self.azimuth, self.wgeom = (
                float(x) for x in f.readline().split()
            )
            self.wavelengths = np.full(self.nconv, np.nan)
            self.spectrum = np.full(self.nconv, np.nan)
            self.error = np.full(self.nconv, np.nan)
            for i in range(self.nconv):
                w, s, e = (float(x) for x in f.readline().split())
                self.wavelengths[i] = w
                self.spectrum[i] = s
                self.error[i] = e

    def write(self, filepath, exclude_nans=True):
        """
        Write an SPX file to disk

        Args:
            filepath (str): The filepath of the new .spx file

        Returns:
            None
        """

        lines = []
        lines.append(f'{self.fwhm}\t{self.lat}\t{self.lon}\t1')
        if exclude_nans:
            mask = ~(np.isnan(self.spectrum) | np.isnan(self.error))
        else:
            mask = np.ones_like(self.spectrum).astype(bool)
        nconv = len(self.wavelengths[mask])
        lines.append(f'{nconv}')
        lines.append(f'{self.nav}')
        lines.append(f'{self.lat}\t{self.lon}\t{self.phase}\t{self.emission}\t{self.azimuth}\t{self.wgeom}')

        if np.max(self.spectrum) > 1e-4:
            warnings.warn(f"This spectrum looks extremely bright (max value {np.max(self.spectrum)}) - have you converted from MJy/sr to W/cm2/sr/um?")

        for w, s, e in zip(self.wavelengths[mask], self.spectrum[mask], self.error[mask]):
            lines.append(f'{w: 12f}  {s: 12e}  {e: 12e}')
        lines.append('')  # end file with a newline

        with open(filepath, 'w') as f:
            f.write('\n'.join(lines))

    @property
    def wavelength(self):
        return self.wavelengths
    
    @wavelength.setter
    def wavelength(self, value):
        self.wavelengths = np.array(value)
    
    @property
    def errors(self):
        return self.error
    
    @errors.setter
    def errors(self, value):
        self.error = np.array(value)

    @classmethod
    def from_lists(cls,
                   wavelengths, 
                   spectrum, 
                   error, 
                   lat, lon, phase, emission, azimuth, fwhm=0,
                   convert=True):
        """Create a new NemesisSpx object from three lists (wavelength, spectrum, errors) and a set of parameters (lat, lon etc...)
        See also: `NemesisSpx.get_angles()`

        Args:
            wavelengths (list or array-like): Wavelength values of the observation.
            spectrum (list or array-like): Measured spectral radiance/reflectance values in units of MJy/sr or W/cm2/sr/um (see `convert`).
            error (list or array-like): Measurement uncertainties for each spectral point.
            lat (float): Latitude of the observation point (degrees).
            lon (float): Longitude of the observation point (degrees).
            phase (float): Phase angle between the observer, target, and light source (degrees).
            emission (float): Emission angle of the observation (degrees).
            azimuth (float): Azimuth angle (degrees).
            fwhm (float, optional): Full width at half maximum of the spectral resolution. Defaults to 0.
            convert (bool, optional): If True, converts the input values from MJy/sr to W/cm2/sr/um.
                                    Defaults to False.

        Returns:
            NemesisSpx: A new instance populated with the given spectral and geometric data.
"""
        spx = cls.__new__(cls)
        if convert:
            spx.spectrum = NemesisSpx.convert_MJysr_to_Wcm2srum(wavelengths, spectrum)
            spx.error = NemesisSpx.convert_MJysr_to_Wcm2srum(wavelengths, error)
        else:
            spx.spectrum = np.array(spectrum)
            spx.error = np.array(error)
        spx.wavelengths = np.array(wavelengths)
        spx.lat = lat
        spx.lon = lon
        spx.phase = phase
        spx.emission = emission
        spx.azimuth = azimuth
        spx.fwhm = fwhm
        spx.nconv = len(spx.wavelengths)
        spx.nav = 1
        spx.wgeom = 1
        return spx

    def get_angles(self):
        """
        Retrieve the geometric parameters of the observation. useful for creating modified versions of the same spectrum (see `NemesisSpx.from_lists()`)

        Args:
            None

        Returns:
            dict: Dictionary containing:
                - 'lat' (float): Latitude in degrees.
                - 'lon' (float): Longitude in degrees.
                - 'phase' (float): Phase angle in degrees.
                - 'emission' (float): Emission angle in degrees.
                - 'azimuth' (float): Azimuth angle in degrees.
        """
        return {
            "lat": self.lat,
            "lon": self.lon,
            "phase": self.phase,
            "emission": self.emission,
            "azimuth": self.azimuth
        }

    @staticmethod
    def convert_MJysr_to_Wcm2srum(
        wavelengths: np.ndarray,
        spectrum: np.ndarray,
    ) -> np.ndarray:
        """
        Convert a spectrum or cube from MJy/sr to W/cm2/sr/micron.
        """
        while len(wavelengths.shape) < len(spectrum.shape):
            # work for cubes passed to sp
            wavelengths = np.expand_dims(wavelengths, -1)

        spx_MJysr = spectrum * u.MJy / u.sr
        spx_Wcm2srum = spx_MJysr.to(
            u.W / (u.cm * u.cm) / u.sr / u.micron,
            equivalencies=u.spectral_density(wavelengths * u.micron),
        )

        return spx_Wcm2srum.value

    @staticmethod
    def convert_Wcm2srum_to_MJysr(
        wavelengths: np.ndarray,
        spectrum: np.ndarray,
    ) -> np.ndarray:
        """
        Convert a spectrum or cube from W/cm2/sr/micron to MJy/sr.
        """
        while len(wavelengths.shape) < len(spectrum.shape):
            # work for cubes passed to sp
            wavelengths = np.expand_dims(wavelengths, -1)

        spx_Wcm2srum = spectrum * u.W / (u.cm * u.cm) / u.sr / u.micron
        spx_MJysr = spx_Wcm2srum.to(
            u.MJy / u.sr,
            equivalencies=u.spectral_density(wavelengths * u.micron),
        )
        return spx_MJysr.value


class NemesisInp(Parser):
    """Parser for nemesis.inp
    
    Attributes:
        wavelength (bool):      Whether to use wavelength or wavenumber
        scattering (bool):      Whether to use a scattering run
        linebyline (bool):      Whether to use line-by-line or corrrelated-k method
        woff (float):           Wavelength offset
        fmerror_path (str):     Path to the file contianing forward modelling errors for each wavelength
        num_iterations (int):   Maximum number of iterations 
        min_phi_change (float): Minimim percentage change in phi before terminating
        num_spectra (int):      Number of spectra to retrieve
        start_spectra (int):    ID of spectra to start with
        sub_previous (bool):    Whether to use a previous retrieval
        output_format (int):    Format of the output files
    """
    
    def read(self):
        with open(self.filepath) as file:
            lines = file.read().split("\n")
            self.wavelength, self.scattering, self.linebyline = map(bool, utils.get_ints_from_string(lines[0]))
            self.woff, = utils.get_floats_from_string(lines[1])
            self.fmerror_path = lines[2]
            self.num_iterations, = utils.get_ints_from_string(lines[3])
            self.min_phi_change, = utils.get_floats_from_string(lines[4])
            self.num_spectra, self.start_spectra = utils.get_ints_from_string(lines[5])
            self.sub_previous, = map(bool, utils.get_ints_from_string(lines[6]))
            self.output_format, = utils.get_ints_from_string(lines[7])


class MakephaseOut(Parser):
    def __init__(self, filepath):
        super().__init__(filepath)
        self.add_aerosol_names()

    def read(self):
        xsc = NemesisXsc(Path(self.filepath).parent / "nemesis.xsc")
        wavelengths = xsc.xsc.wavelength

        with open(self.filepath) as file:
            isprev = False
            start = False
            out = []
            for line in file:
                if not isprev and not start:
                    out.append([])
                    start = True
                if line.startswith(" Refractive index:"):
                    isprev = True
                    start = False
                    out[-1].append(utils.get_floats_from_string(line))
                else:
                    isprev = False
        
        out = pd.DataFrame(np.hstack(out))
        out.insert(0, "wavelength", wavelengths)
        self.data = out

    def add_aerosol_names(self):
        # Get the names of the aerosol layers
        names = []
        with open(Path(self.filepath).parent / "aerosol_names.txt") as file:
            n = file.read().split("\n")
            for m in n:
                names.append(m + " real")
                names.append(m + " imag")
        
   
class AerosolPrf(Parser):
    """Parser for the aerosol.prf file
    
    Attributes:
        data: pd.DataFrame containing the aerosol density as a function of height. The units of aerosol density are particles per gram of atmosphere"""
    
    def __init__(self, filepath):
        super().__init__(filepath)
        self.add_aerosol_names()

    def read(self):
        self.data = pd.read_table(self.filepath, sep="\s+", skiprows=2, header=None)
        num = len(self.data.columns) - 1
        header = ["height"] + [f"aerosol_{x}" for x in range(1, num+1)]
        self.data.columns = header

    def add_aerosol_names(self):
        with open(Path(self.filepath).parent / "aerosol_names.txt") as file:
            names = file.read().split("\n")
            if names == [""]:
                return
            self.data.columns = ["height"] + names


class ParaH2Ref(Parser):
    """Parser for the parah2.ref file
    
    Attributes:
        data: pd.DataFrame containing the aerosol density as a function of height. The units of aerosol density are particles per gram of atmosphere"""
    
    def __init__(self, filepath):
        super().__init__(filepath)

    def read(self):
        self.data = pd.read_table(self.filepath, sep="\s+", skiprows=1, header=None)
        header = ["height", "parah2_fraction"]
        self.data.columns = header


class kTable(Parser):
    """Parser for ktables
    
    Atributes:
        ktable (np.ndarray):      4D array of absorption coefficients with axes (wavelength, pressure, temperature, g-ordinate).
        wavelength (np.ndarray):  Array of spectral points
        pressure (np.ndarray):    Array of pressure levels
        temperature (np.ndarray): Array of temperature levels
        g_ordinates (np.ndarray): Array of g-ordinates used in correlated-k integration
        g_weights (np.ndarray):   Corresponding quadrature weights for g-ordinates
        irec0 (int):              Number of header records
        npoint (int):             Number of spectral points
        vmin (float):             Starting value of the spectral axis
        delv (float):             Spectral spacing. If negative, grid is non-uniform and read from file.
        fwhm (float):             Full Width at Half Maximum for spectral bins
        np_ (int):                Number of pressure levels
        nt (int):                 Number of temperature levels
        ng (int):                 Number of g-ordinates
        idgas (int):              Identifier for the gas species 
        isogas (int):             Identifier for the isotope of the gas
    """
    
    @staticmethod
    def read_float(file):
        return struct.unpack('f', file.read(4))[0]

    @staticmethod
    def read_long(file):
        return struct.unpack('i', file.read(4))[0]

    def read(self):
        with open(self.filepath, 'rb') as f:

            # Read header
            self.irec0  = self.read_long(f)
            self.npoint = self.read_long(f)
            self.vmin   = self.read_float(f)
            self.delv   = self.read_float(f)
            self.fwhm   = self.read_float(f)
            self.np_    = self.read_long(f)
            self.nt     = self.read_long(f)
            self.ng     = self.read_long(f)
            self.idgas  = self.read_long(f)
            self.isogas = self.read_long(f)

            # Read g ordinates and weights
            self.g_ordinates = np.array([self.read_float(f) for _ in range(self.ng)])
            self.g_weights = np.array([self.read_float(f) for _ in range(self.ng)])

            # Skip two records
            _ = self.read_float(f)
            _ = self.read_float(f)

            # Construct pressure/temperature grid
            self.pressure = np.array([self.read_float(f) for _ in range(self.np_)])
            self.temperature = np.array([self.read_float(f) for _ in range(self.nt)])

            # Read in wavelengths
            nrec = 10 + 2 * self.ng + 2 + self.np_ + self.nt
            if self.delv < 0.0:
                self.wavelength = np.array([self.read_float(f) for _ in range(self.npoint)])
                nrec += self.npoint
            else:
                self.wavelength = self.vmin + self.delv * np.arange(self.npoint)

            # Skip records to reach k-table data
            jrec = self.irec0 - nrec - 1
            for _ in range(jrec):
                _ = self.read_float(f)

            # Read k-table
            total_values = self.npoint * self.np_ * self.nt * self.ng
            ktable_flat = np.frombuffer(f.read(total_values * 4), dtype=np.float32)
            self.ktable = ktable_flat.reshape((self.npoint, self.np_, self.nt, self.ng))


class PsgSpectrum(Parser):
    def read(self):
        self.data = pd.read_csv(self.filepath, comment="#", header=None, sep="\s+")
        self.data.columns = ["wavelength", "spectrum", "drop"]
        self.data.drop(["drop"], axis=1, inplace=True)


class NemesisCov(Parser):
    """Parser for nemesis.cov
    
    Attributes:
        cov (np.ndarray): Covariance matrix of the retrieved parameters"""
    
    def read(self):
        raise NotImplementedError("This parser is not finished. Suppress this error at your peril")
        with open(self.filepath, 'r') as f:
            lines = f.readlines()

        _, nvar = utils.get_ints_from_string(lines[0])
        nx, ny = utils.get_ints_from_string(lines[2*nvar + 1])
        start = 2 * nvar + 2

        sa = [utils.get_floats_from_string(lines[start + i]) for i in range(nx)]
        self.sa = np.array(sa)
 


""" 
Each Shape object represents a differnet model in NEMESIS. Eg. ``Shape1`` represents profile shape 1 which is a constant VMR 
up to a knee pressure then a fractional scale height. For a ful list of possible models see the NEMESIS manual. Please note: 
due to the large number of profile shapes NEMESIS supports, 

Adding a Shape subclass
=======================

All subclasses of Shape must implement:
    - ID: Integer ID of the shape in NEMESIS
    - CONSTANTS: The names of parameters that are not fitted by NEMESIS but required for the model
    - VARIABLES: The names of the parameters in order they appear in the mre file (do not include error parameters here). **These MUST be defined in the order that they appear in the .apr file!!**
    - generate_apr_data: Returns a string of the profile shape parameters in the format
                         that the .apr file expects.

Additionally, some shapes require additional files in the core (eg. tempapr.dat). If this is the case,
override the create_required_files method.

Finally, add the new ShapeN class to the ALL_SHAPES list at the end of the file.

Class template::

    @shapeclass
    class ShapeN(Shape):
        "copy the desciption from the NEMESIS manual here (use triple quotes!)"
        ID: ClassVar[int] = N
        CONSTANTS: ClassVar[list[str]] = ["arg_name_1", ...]
        VARIABLES: ClassVar[list[str]] = ["arg_name_2", "arg_name_3", ...]
        arg_name_1: type_1
        arg_name_2: type_2
        arg_name_2_error: type_2
        ...

        def generate_apr_data(self):
            return a string for apr file

        # optionally
        def create_required_files(self, directory):
            shutil.copy(..., directory)

"""

import numpy as np

from dataclasses import dataclass
from typing import ClassVar
import shutil

from . import parsers
from . import utils


def shapeclass(*args, **kwargs):
    """Syntatic sugar for ``@dataclass(repr=False)``"""
    return dataclass(*args, **kwargs, repr=False)
    


class Shape:
    def __init__(self):
        """Do not instantiate directly, use a subclass"""
        pass

    def __repr__(self):
        return f"<Shape{self.ID}>"
    
    def _set_prior_to_retrieved(self):
        """Set the prior values to the retrieved values and delete the retrived_* attributes."""
        for base in self.VARIABLES:
            for name in [base, f"{base}_error"]:
                try:
                    setattr(self, name, getattr(self, "retrieved_"+name))
                    delattr(self, "retrieved_"+name)
                except AttributeError:
                    print(f"Clear attribute failed for {name} on {self}")

    def generate_apr_data(self):
        """Generate the section of the .apr file that stores the shape parameters
        
        Args:
            None
            
        Returns:
            str: The string to write to the .apr file"""
        pass

    def create_required_files(self, directory):
        """Some Shapes require additional files to be created/copied into the core directory"""
        pass

    def sample_from_distribution(self, **kwargs):
        """
        Set parameters of the Shape to random values pulled from a normal distribution, with arguments mean_[parameter] and std_[parameter].
        For example: calling Shape4.sample_from_distribution(mean_deep_vmr=0.1, std_deep_vmr=0.05) will set the deep vmr to a random value drawn 
        from a gaussian with the centre at 0.1 and a standard deviation of 0.05. Multiple parameters may be specified at once.
        
        Args:
            * mean_[parameter]: The mean of the distribution to sample for parameter [parameter]
            * std_[parameter]:  The standard deviation of the distribution to sample for parameter [parameter]
        
        Returns:
            None
        """

        for param in kwargs.keys():
            for vp in self.CONSTANTS + self.VARIABLES:
                if vp in param:
                    if f"mean_{vp}" in kwargs.keys():
                        mean = kwargs[f"mean_vp"]
                    else:
                        mean = getattr(self, vp)
                    setattr(self, vp, np.random.normal(mean, kwargs[f"std_{vp}"]))
        


@shapeclass
class Shape0(Shape):
    """Profile is to be treated as continuous over the pressure range of runname.ref, the
    next line of the .apr file should then contain a filename, which specifies the a
    priori profile as a function of height and should have the same number of levels
    as the .ref file.
    
    In Eleos, one can specify either a filepath or give the values and error to be used directly.

    Parameters
    ----------
    filepath : str, optional
        Path to the input profile file.
    values : list[float], optional
        Profile values to be written if no file is given.
    errors : list[float], optional
        Error values corresponding to `values`.
    """
    ID: ClassVar[int] = 0
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = []
    filepath: str = None
    values: list[float] = None
    errors: list[float] = None

    def create_required_files(self, directory):
        if self.filepath is not None:
            shutil.copy(self.filepath, directory)
            self.values = []
            self.errors = []
            with open(self.filepath) as file:
                lines = file.read().split("\n")[1:]
                for line in lines:
                    p, v, e = utils.get_floats_from_string(line)
                    self.values.append(v)
                    self.errors.append(e)
        else:
            self.filepath = directory / "shape0.dat"
            ref = parsers.NemesisRef(directory / "nemesis.ref")
            with open(self.filepath, mode="w+") as file:
                file.write(f"{len(ref.data.pressure)} 1\n#")
                for p, v, e in zip(ref.data.pressure, self.values, self.errors):
                    file.write(f"{p:08e}  {v:08e}  {e:08e}\n")

    def generate_apr_data(self):
        return self.filepath.name

@shapeclass
class Shape1(Shape):
    """Profile is to be represented as a deep VMR up to a certain ‘knee’ pressure, and
    then a defined fractional scale height.

    Parameters
    ----------
    knee_pressure : float
        Pressure at which the transition occurs.
    deep_vmr : float
        Deep volume mixing ratio.
    deep_vmr_error : float
        Error estimate for deep_vmr.
    fsh : float
        Fractional scale height.
    fsh_error : float
        Error estimate for fsh.
    """
    ID: ClassVar[int] = 1
    CONSTANTS: ClassVar[list[str]] = ["knee_pressure"]
    VARIABLES: ClassVar[list[str]] = ["deep_vmr", "fsh"]
    knee_pressure: float
    deep_vmr: float
    deep_vmr_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.knee_pressure}\n{self.deep_vmr} {self.deep_vmr_error}\n{self.fsh} {self.fsh_error}"

@shapeclass
class Shape2(Shape):
    """Profile is to be represented by a simple scaling of the corresponding profile
    runname.ref, aerosol.ref, parah2.ref or fcloud.ref.

    Parameters
    ----------
    scale_factor : float
        Scaling factor applied to the reference profile.
    scale_factor_error : float
        Error estimate for scale_factor.
    """
    ID: ClassVar[int] = 2
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = ["scale_factor"]
    scale_factor: float
    scale_factor_error: float

    def generate_apr_data(self):
        return f"{self.scale_factor} {self.scale_factor_error}"

@shapeclass
class Shape4(Shape):
    """Very similar to Shape1, but knee pressure is also a variable parameter.

    Parameters
    ----------
    knee_pressure : float
        Pressure at which the transition occurs.
    knee_pressure_error : float
        Error estimate for knee_pressure.
    deep_vmr : float
        Deep volume mixing ratio.
    deep_vmr_error : float
        Error estimate for deep_vmr.
    fsh : float
        Fractional scale height.
    fsh_error : float
        Error estimate for fsh.
    """
    ID: ClassVar[int] = 4
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = ["deep_vmr", "fsh", "knee_pressure"]
    knee_pressure: float
    knee_pressure_error: float
    deep_vmr: float
    deep_vmr_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.knee_pressure} {self.knee_pressure_error}\n{self.deep_vmr} {self.deep_vmr_error}\n{self.fsh} {self.fsh_error}"

@shapeclass
class Shape20(Shape):
    """Similar to Shape1, but profile forced to a small number above tropopause.

    Parameters
    ----------
    knee_pressure : float
        Pressure at which the transition occurs.
    tropopause_pressure : float
        Pressure representing the tropopause.
    deep_vmr : float
        Deep volume mixing ratio.
    deep_vmr_error : float
        Error estimate for deep_vmr.
    fsh : float
        Fractional scale height.
    fsh_error : float
        Error estimate for fsh.
    """
    ID: ClassVar[int] = 20
    CONSTANTS: ClassVar[list[str]] = ["knee_pressure", "tropopause_pressure"]
    VARIABLES: ClassVar[list[str]] = ["deep_vmr", "fsh"]
    knee_pressure: float
    tropopause_pressure: float
    deep_vmr: float
    deep_vmr_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.knee_pressure} {self.tropopause_pressure}\n{self.deep_vmr} {self.deep_vmr_error}\n{self.fsh} {self.fsh_error}"

@shapeclass
class Shape32(Shape):
    """Cloud profile with variable base pressure, opacity, and fractional scale height.

    Parameters
    ----------
    base_pressure : float
        Cloud base pressure (bar).
    base_pressure_error : float
        Error estimate for base_pressure.
    opacity : float
        Cloud opacity (log-scaled).
    opacity_error : float
        Error estimate for opacity.
    fsh : float
        Fractional scale height (log-scaled).
    fsh_error : float
        Error estimate for fsh.
    """
    ID:  ClassVar[int] = 32
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = ["opacity", "fsh", "base_pressure"]
    base_pressure: float
    base_pressure_error: float
    opacity: float
    opacity_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.base_pressure} {self.base_pressure_error}\n{self.opacity} {self.opacity_error}\n{self.fsh} {self.fsh_error}"
        
@shapeclass
class Shape37(Shape):
    """Cloud with constant opacity/bar between two pressure levels.

    Parameters
    ----------
    bottom_pressure : float
        Bottom pressure level (bar).
    top_pressure : float
        Top pressure level (bar).
    opacity : float
        Cloud opacity per bar.
    opacity_error : float
        Error estimate for opacity.
    """
    ID: ClassVar[int] = 37
    CONSTANTS: ClassVar[list[str]] = ["bottom_pressure", "top_pressure"]
    VARIABLES: ClassVar[list[str]] = ["opacity"]
    bottom_pressure: float
    top_pressure: float
    opacity: float
    opacity_error: float

    def generate_apr_data(self):
        return f"{self.bottom_pressure} {self.top_pressure}\n{self.opacity} {self.opacity_error}"

@shapeclass
class Shape47(Shape):
    """Cloud centred at specified pressure with variable FWHM and opacity.

    Parameters
    ----------
    central_pressure : float
        Pressure where cloud distribution peaks (bar).
    central_pressure_error : float
        Error estimate for central_pressure (bar).
    pressure_width : float
        Full width at half maximum in log pressure units.
    pressure_width_error : float
        Error estimate for pressure_width.
    opacity : float
        Total cloud opacity.
    opacity_error : float
        Error estimate for opacity.
    """
    ID: ClassVar[int] = 47
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = ["opacity", "central_pressure", "pressure_width"]
    central_pressure: float
    central_pressure_error: float
    pressure_width: float
    pressure_width_error: float
    opacity: float
    opacity_error: float

    def generate_apr_data(self):
        return f"{self.opacity} {self.opacity_error}\n{self.central_pressure} {self.central_pressure_error}\n{self.pressure_width} {self.pressure_width_error}"

@shapeclass
class Shape48(Shape):
    """Cloud profile with variable base, top pressure, opacity, and scale height.

    Parameters
    ----------
    base_pressure : float
        Cloud base pressure (bar).
    base_pressure_error : float
        Error estimate for base_pressure (bar).
    top_pressure : float
        Cloud top pressure (bar).
    top_pressure_error : float
        Error estimate for top_pressure (bar).
    opacity : float
        Cloud opacity.
    opacity_error : float
        Error estimate for opacity.
    fsh : float
        Fractional scale height.
    fsh_error : float
        Error estimate for fsh.
    """
    ID:  ClassVar[int] = 48
    CONSTANTS: ClassVar[list[str]] = []
    VARIABLES: ClassVar[list[str]] = ["opacity", "fsh", "base_pressure", "top_pressure"]
    base_pressure: float
    base_pressure_error: float
    top_pressure: float
    top_pressure_error: float
    opacity: float
    opacity_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.base_pressure} {self.base_pressure_error}\n{self.top_pressure} {self.top_pressure_error}\n{self.opacity} {self.opacity_error}\n{self.fsh} {self.fsh_error}"


def get_shape_from_id(id_):
    """Given a shape ID integer, return a reference to the class corresponding to that ID. Note that this returns a class, not an instantiated object.
    
    Args:
        id_: Shape ID to get
        
    Returns:
        Shape: Class object of the corresponding shape"""
    id_ = int(id_)
    for shape_class in ALL_SHAPES:
        if shape_class.ID == id_:
            return shape_class


ALL_SHAPES = [Shape0, Shape1, Shape2, Shape4, Shape20, Shape32, Shape37, Shape47, Shape48] #: A list of all the Shape classes


import math
import pickle
import time
import sys
import signal
import os
import shutil
import re
import copy
import subprocess
from collections import defaultdict
from pathlib import Path
import pandas as pd
import numpy as np

import warnings
warnings.formatwarning = lambda msg, *_: f"Warning: {msg}\n"

from . import constants
from . import spx
from . import parsers
from . import utils
from . import profiles as profiles_ # to avoid namespace shadowing by NemesisCore.profiles
from . import results

CORE_ID_COUNTER = 0


class NemesisCore:
    """The class that constructs a core directory and all the required files for NEMESIS to run.
    
    Attributes:
        parent_directory (Path):      The directory in which the core directory is created
        directory (Path):             The directory in which the core files are stored
        id_ (int):                    The unique ID of the core
        profiles (dict):              A dictionary of profiles.Profile objects to retrieve, where the keys are the profile labels and the values are the Profile objects themselves
        spx_file (Path):              The path to the spectrum file to fit
        ref_file (Path):              The path to the ref file to use
        cia_file (Path):              The path to the collision-induced absorbtion file to use (default: mgRT_mgRV_40-400K_dnu4.0.tab, warning: always assumes d nu = 4.0)
        sol_file (Path):              The path to the solar spectrum file to use (default: solar_spec.dat)
        ref (parsers.NemesisRef):     The parsed ref file
        planet (str):                 The name of the planet being observed
        scattering (bool):            Whether to run a scattering retrieval or not
        num_aerosol_modes (int):      The number of aerosol profiles in the retrieval
        forward (bool):               Whether of not to run a forward model (ie. set number of iterations = 0)
        num_iterations (int):         Number of iterations to run in the retrieval (if forward is set this has no effect)
        num_layers (int):             The number of atmospheric layers to simulate
        layer_type (int):             The type of layering to use 
        min_pressure (float):         The minimum pressure in the atmosphere
        max_pressure (float):         The maximum pressure in the atmosphere
        bottom_layer_height (int):    The height in km of the bottom of the atmosphere
        fixed_peaks (list):           A list of FixedPeak objects that specify regions of the spectrum to fix
        instrument_ktables (str):     Either 'NIRSPEC' or 'MIRI'; determines which set of ktables to use.
        fmerror_factor (float):       The factor by which to multiply the error on the spectrum (see also, fmerror_pct and fmerror_value)
        fmerror_pct (float):          If given, instead of using fmerror_factor or fmerror_value, use a flat percentage of the brightness (eg. 0.1 = 10%) (see also, fmerror_factor and fmerror_value)
        fmerror_value (float):        If given, instead of using fmerror_factor or fmerror_pct, use a flat value in W/cm2/sr/um (see also, fmerror_factor and fmerror_pct)
        cloud_cover (bool):           If scattering mode is on, then this is the fractional cloud cover between 0 and 1 (usually doesn't need to be changed)
        reference_wavelength (float): If scattering mode is on, then normalise the cross-sections at the closest wavelength to this value in the .xsc file
        """
    
    def __init__(self, 
                 parent_directory,
                 spx_file, 
                 ref_file=None, 
                 cia_file=None,
                 sol_file=None,
                 profiles=list(), 
                 planet="jupiter", 
                 scattering=True, 
                 forward=False,
                 num_iterations=20,
                 num_layers=120, 
                 min_pressure=None,
                 max_pressure=None,
                 instrument_ktables="NIRSPEC", 
                 fmerror_factor=None,
                 fmerror_pct=None,
                 fmerror_value=None,
                 cloud_cover=1.0,
                 reference_wavelength=None):
        
        """Create a NEMESIS core directory with a given set of profiles to retrieve
        
        Args:
            parent_directory (str):       The directory in which to create the core folder 
            spx_file (str):               Path to the spectrum file to fit
            ref_file (str):               Path to the ref file to use (if left blank it will use the default for that planet)
            profiles List[str]:           List of profiles.Profile objects to retrieve
            planet (str):                 Name of the planet being observed. Must be one of 'jupiter', 'saturn', 'uranus', 'neptune' or 'titan'.
            scattering (bool):            Whether to run a scattering retrieval or not
            forward (bool):               Whether of not to run a forward model (ie. set number of iterations = 0)
            num_iterations (int):         Number of iterations to run in the retrieval (if forward is set this has no effect)
            num_layers (int):             The number of atmospheric layers to simulate
            min_pressure (float):         The pressure at the top of the model atmosphere
            max_pressure (float):         The pressure at the bottom of the model atmosphere
            instrument_ktables (str):     Either 'NIRSPEC' or 'MIRI'; determines which set of ktables to use.
            fmerror_factor (float):       The factor by which to multiply the error on the spectrum (see also, fmerror_pct and fmerror_value)
            fmerror_pct (float):          If given, instead of using fmerror_factor or fmerror_value, use a flat percentage of the brightness (eg. 0.1 = 10%) (see also, fmerror_factor and fmerror_value)
            fmerror_value (float):        If given, instead of using fmerror_factor or fmerror_pct, use a flat value in W/cm2/sr/um (see also, fmerror_factor and fmerror_pct)
            cloud_cover (bool):           If scattering mode is on, then this is the fractional cloud cover between 0 and 1 (usually doesn't need to be changed)
            reference_wavelength (float): If scattering mode is on, then normalise the cross-sections at the closest wavelength to this value in the .xsc file. This parameter will be updated with the exact wavelength
        """
        # Increment the global core counter and store local version
        global CORE_ID_COUNTER
        CORE_ID_COUNTER += 1
        self.id_ = CORE_ID_COUNTER

        # Assign attributes passed in
        self.spx_file = Path(spx_file).resolve()
        self.planet = planet.lower()
        self.scattering = scattering
        self._forward = forward
        self.num_iterations = num_iterations
        self.num_layers = num_layers
        self.fmerror_factor = fmerror_factor
        self.fmerror_pct = fmerror_pct
        self.fmerror_value = fmerror_value
        self.cloud_cover = cloud_cover
        self.reference_wavelength = reference_wavelength

        # Set up instrument ktables
        if instrument_ktables.lower() not in ["nirspec", "miri"]:
            raise ValueError(f"instrument_ktables must be either 'NIRSPEC' or 'MIRI', not {instrument_ktables}")
        self.instrument_ktables = instrument_ktables
        self.ktable_path = constants.PATH / f"data/jupiter/{self.instrument_ktables.lower()}.kls"

        # Set the directories of the parent folder and own core
        self.parent_directory = Path(parent_directory).resolve()
        self.directory = self.parent_directory / f"core_{self.id_}"

        # Check if we are perfoming a scattering run with more than 39 layers (NEMESIS hates this...)
        if num_layers > 39 and self.scattering:
            warnings.warn(f"Too many atmospheric layers specified for a scattering run ({num_layers} vs. 39). Automatically reducing to 39")
            self.num_layers = 39

        # Set ref file if not specified:
        if ref_file is None:
            self.ref_file = constants.PATH / f"data/{planet}/{planet}.ref"
            warnings.warn(f"No ref file specified. Using the default in {self.ref_file}")
        else:
            self.ref_file = ref_file

        # Set cia file if not set
        if cia_file is not None:
            # if the file can be found from here then it is not in raddata, so add it
            if Path(cia_file).exists():
                add_to_raddata(cia_file)
                
            # If it isnt in raddata either, raise an error
            elif not (constants.NEMESIS_PATH / "radtrancode/raddata" / Path(cia_file)).exists():
                raise FileNotFoundError(f"Could not find {cia_file} here or in raddata/")
            self.cia_file = Path(cia_file).name
        else:
            self.cia_file = Path("mgRT_mgRV_40-400K_dnu4.0.tab")

        # Set sol file if not set
        if sol_file is not None:
            sol_file = Path(sol_file)
            # if the file can be found from here then it is not in raddata, so add it
            if sol_file.exists():
                add_to_raddata(Path(sol_file))
                
            # If it isnt in raddata either, raise an error
            elif not (constants.NEMESIS_PATH / "radtrancode/raddata" / sol_file).exists():
                raise FileNotFoundError(f"Could not find {sol_file} here or in raddata/")
            self.sol_file = sol_file 
        else:
            self.sol_file = Path("solar_spec.dat")

        # Parse the ref file:
        self.ref = parsers.NemesisRef(self.ref_file)

        # Set min/max pressures
        self.set_pressure_limits(max_pressure=max_pressure, 
                                 min_pressure=min_pressure)

        # Set number of aerosol modes (incremented by add_profile)
        self.num_aerosol_modes = 0

        # Add a list to hold any FixedPeak classes the user wants
        self.fixed_peaks = []
        self.excluded_gases = []

        # Add a reference to self in each Profile and set up AerosolProfiles correctly
        self.profiles = dict()
        for profile in profiles:
            self.add_profile(profile)

        # If in forward mode, set the number of iterations to 0
        if self.forward:
            self.num_iterations = 0

        # Raise an error if trying to use features not implemented yet
        if planet != "jupiter":
            raise Exception("Eleos does not support planets other than Jupiter yet! ")

    def __str__(self):
        return f"<NemesisCore: {self.directory}>"

    @property
    def forward(self):
        return self._forward
    
    @forward.setter
    def forward(self, value):
        self._forward = value
        if value:
            self.num_iterations = 0
        else:
            self.num_iterations = 20

    def _create_directory_tree(self, prompt_if_exists):
        os.makedirs(self.parent_directory, exist_ok=True)
        if os.path.exists(self.directory) and prompt_if_exists:
            if os.path.exists(self.directory / "nemesis.mre"):
                msg = "There is already a core that has been run in"
            elif os.path.exists(self.directory / "nemesis.ref"):
                msg = "There is already a core that has not been run yet in"
            else:
                msg = "There is already data in"

            x = input(f"{msg} {self.directory.resolve()} - erase and continue? Y/N ")
            if x.lower() == "n":
                return
            shutil.rmtree(self.directory)
        os.makedirs(self.directory / "plots", exist_ok=True)

    def _save_core(self):
        """Dump the core object to a pickle file in the core directory
        
        Args:
            None
        
        Returns:
            None
            
        Creates:
            core.pkl"""
        with open(self.directory / "core.pkl", mode="wb") as file:
            pickle.dump(self, file)

    def _copy_input_files(self):
        """Copy the given .spx and .ref file into the core, as well as the eleos generation script itself
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            nemesis.spx
            nemesis.ref
            eleos_generation.py"""

        shutil.copy(self.ref_file, self.directory / "nemesis.ref")
        shutil.copy(self.spx_file, self.directory / "nemesis.spx")
        shutil.copy(sys.argv[0], self.directory / "eleos_generation.py")

    def _copy_template_files(self):
        """Copy some boilerplate files from the library data directory into the core directory
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            nemesis.abo
            nemesis.nam"""
        shutil.copy(constants.PATH / "data/statics/nemesis.abo", self.directory)
        shutil.copy(constants.PATH / "data/statics/nemesis.nam", self.directory)

    def _generate_cia(self):
        """Generate the .cia (collision induced absorbtion) file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            nemesis.cia
        """
        with open(self.directory / "nemesis.cia", mode="w+") as file:
            file.write(self.cia_file.name + "\n")
            file.write("4.00000\n")
            file.write("24\n")

    def _generate_sol(self):
        """Generate the .sol (solar spectrum) file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            nemesis.cia
        """
        with open(self.directory / "nemesis.sol", mode="w+") as file:
            file.write(self.sol_file.name)

    def _generate_inp(self):
        """Generate the nemesis input file

        Args:
            num_iterations: Number of iterations to run (has no effect if NemesisCore.forward is True)
        
        Returns:
            None
            
        Creates:
            nemesis.inp"""
        with open(constants.PATH / "data/statics/nemesis.inp") as file:
            out = file.read()
            out = out.replace("<SCATTERING>", str(int(self.scattering)))
            out = out.replace("<N_ITERATIONS>", str(self.num_iterations))
        with open(self.directory / "nemesis.inp", mode="w+") as file:
            file.write(out)

    def _generate_apr(self):
        """Generate the nemesis.apr file from the profile list

        Args:
            None
        
        Returns:
            None
            
        Creates:
            nemesis.apr"""
        
        out = f"*******Apriori File*******\n           <NUM_PROFILES>\n"

        num_profiles = 0
        for profile in self.profiles.values():

            profile.shape.create_required_files(self.directory)

            if isinstance(profile, profiles_.AerosolProfile) and profile.retrieve_optical:
                profile._generate_cloudfn_dat(self.directory)
                num_profiles += 1

            out += profile.generate_apr_data() + "\n"
            num_profiles += 1

        with open(self.directory / "nemesis.apr", mode="w") as file:
            out = out.replace("<NUM_PROFILES>", str(num_profiles))
            file.write(out)

    def _generate_set(self):
        """Generate the settings file for NEMESIS
        
        Args:
            None

        Returns:
            None
            
        Creates:
            nemesis.set"""

        with open(constants.PATH / "data/statics/nemesis.set", mode="r") as file:
            out = file.read()
        out = out.replace("<DISTANCE>", f"{constants.DISTANCE[self.planet]:.3f}")
        out = out.replace("<SUNLIGHT>", f"{int(self.scattering)}")
        out = out.replace("<BOUNDARY>", f"{int(self.scattering)}")
        out = out.replace("<N_LAYERS>", f"{int(self.num_layers)}")
        out = out.replace("<BASE>", f"{self.bottom_layer_height:.2f}")
        with open(self.directory / "nemesis.set", mode="w+") as file:
            file.write(out)

    def _generate_flags(self, inormal=0, iray=1, ih2o=0, ich4=0, io3=0, inh3=0, iptf=0, imie=0, iuv=0):
        """Generate the flags file. As a general rule, leave all this at the defaults. The descriptions are copied directly fron the NEMESIS manual.
        Args:
            inormal: 0 or 1 depending on whether the ortho/para-H2 ratio is in equilibrium (0) or normal 3:1 (1).
            iray: Sets the Rayleigh optical depth calculation:
                    0 = Rayleigh scattering optical depth is not included.
                    1 = Rayleigh optical depths suitable for gas giant atmosphere
                    2 = Rayleigh optical depths suitable for CO2 dominated atmosphere
                    3 = Rayleigh optical depths suitable for a N2-O2 atmosphere.
                    4 = New Rayleigh optical depths suitable for a gas giant atmosphere.
                    5 = New Rayleigh optical depths suitable for a gas giant atmosphere PLUS Raman scattering.
                    6 = New Rayleigh optical depths suitable for a gas giant atmosphere PLUS Sromovsky Raman scattering, PLUS Sromovsky Polarisation Correction.
            ih2o: Turns additional H2O continuum off (0) or on (1)
            ich4: Turns additional CH4 continuum off (0) or on (1)
            io3:  Turns additional O3 UV continuum off (0) or on (1)
            inh3: Turns additional NH3 continuum off (0) or on (1)
            iptf: Used in only a few routines to switch between normal partition function calculation (0) or the high-temperature partition function for CH4 for Hot Jupiters.
            imie: Only relevant for scattering calculations. If set to 0, the phase function is computed from the associated Henyey-Greenstein hgphase*.dat files. However, if set to 1, the phase function is computed from the MieTheory calculated PHASEN.DAT file.
        
        Returns:
            None
            
        Creates:
            nemesis.fla"""
        
        with open(self.directory / "nemesis.fla", mode="w+") as file:
            file.write('       {aa}	! Inormal (0=eqm, 1=normal)\n       {bb}	! Iray (0=off, 1=on)\n       {cc}	! IH2O (1 = turn extra continuum on)\n       {dd}	! ICH4 (1 = turn extra continuum on)\n       {ee}	! IO3 (1 = turn extra continuum on)\n       {ff}	! INH3 (1 = turn extra continuum on)\n       {gg}	! Iptf (0=default, 1=CH4 High-T)\n       {hh}	! IMie\n       {ii}	! UV Cross-sections\n       0	! INLTE (0=LTE)'.format(aa=inormal, bb=iray, cc=ih2o, dd=ich4, ee=io3, ff=inh3, gg=iptf, hh=imie, ii=iuv))

    def _generate_aerosol_ref(self):
        """Generate a number of aerosol reference profile that is 0 at all altitudes.
        
        Args:
            None

        Returns:
            None
            
        Creates:
            aerosol.ref"""
        
        heights = self.ref.data.height
        with open(self.directory / "aerosol.ref", mode="w+") as file:
            file.write(f"# Generated by Eleos\n")
            file.write(f"{len(heights)} {self.num_aerosol_modes}\n")
            for h in heights:
                file.write(f"{h:>12.5f} ")
                for i in range(max(self.num_aerosol_modes, 1)):
                    file.write("0.000000E+00 ")
                file.write("\n")

    def _generate_kls(self):
        """Copy the ktables from the template core for the given instrument.
        Args:
            instrument: Either 'NIRSPEC' or 'MIRI'
            
        Returns:
            None
            
        Creates:
            nemesis.kls
            
        TODO:
            Add way to include/exclude different elements
            Add option to use ktables on ALICE rather than prepackaged"""

        mask = [name in self.excluded_gases for name in self.get_ktable_gas_names()]

        with open(self.ktable_path, mode="r") as tmp:
            with open(self.directory / "nemesis.kls", mode="w+") as file:
                for i, line in enumerate(tmp):
                    if not mask[i]:
                        file.write(line)

    def _generate_fmerror(self):
        """For each wavelength in the spx file, adjust the error by a factor (either fmerror_factor, _pct or _value) and write to the fmerror file. 
        Currently only supports 1 spx geometry
        Args:
            None
            
        Returns:
            None
        
        Creates:
            fmerror.dat """
        spx_data = spx.read(self.spx_file).geometries[0]
        num_entries = len(spx_data.wavelengths)

        x = (self.fmerror_factor is not None, self.fmerror_pct is not None, self.fmerror_value is not None)
        if sum(x) > 1:
            raise ValueError("Must specify exactly one of fmerror_factor, fmerror_pct or fmerror_value")

        if num_entries > 2048:
            warnings.warn(f"spx file has too many wavelengths for default NEMESIS installation ({num_entries}/2048) - have you increased IDIM in arrdef.f?")
        
        with open(self.directory / "fmerror.dat", mode="w+") as file:
            # Header with number of lines 
            file.write(f"{num_entries+2}\n")

            # Catch any cases outside the wavelength range (lower)
            file.write(f"{0.1:.6e}  {0:.6e}\n")

            # Scale the spx error by either fmerror_factor, fmerror_pct or fmerror_value
            for wl, err, spc in zip(spx_data.wavelengths, spx_data.error, spx_data.spectrum):

                # Check to see if this peak is fixed
                flag = False
                for fxp in self.fixed_peaks:
                    if fxp.isin(wl):
                        file.write(f"{wl:.6e}  {fxp.error:.6e}\n")
                        flag = True

                # Otherwise, compute the new error
                if self.fmerror_factor is not None and not flag:
                    file.write(f"{wl:.6e}  {err*self.fmerror_factor:.6e}\n")
                if self.fmerror_pct is not None and not flag:
                    file.write(f"{wl:.6e}  {spc*self.fmerror_pct:.6e}\n")
                if self.fmerror_value is not None and not flag:
                    file.write(f"{wl:.6e}  {self.fmerror_value:.6e}\n")

            # Catch any cases outside the wavelength range (upper)
            file.write(f"{100:.6e}  {0:.6e}\n")

    def _generate_xsc(self):
        """Run Makephase and Normxsc with the given inputs to generate the .xsc and phase files.
        
        Args:
            None
            
        Returns:
            None
        
        Creates:
            nemesis.xsc
            PHASE{N}.DAT
            hgphase{n}.dat"""
        
        # # Check to see if any aerosols have been added - if not then return immediately
        # if self.num_aerosol_modes == 0:
        #     return

        # Run Makephase and Normxsc
        cwd = os.getcwd()
        os.chdir(self.directory)
        os.system("Makephase < makephase.inp > makephase.out")
        os.system("Normxsc < normxsc.inp > normxsc.out")
        os.chdir(cwd)

    def _generate_fcloud_ref(self):
        """Generate the fcloud.ref file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            fcloud.ref"""
        
        # Return immediately if there are no cloud layers defined
        if self.num_aerosol_modes == 0:
            return 
        
        with open(self.directory / "fcloud.ref", mode="w+") as file:
            file.write(f"{len(self.ref.data)}    {self.num_aerosol_modes}\n")
            for height in self.ref.data.height:
                file.write(f"      {height:> 9.4f}      {int(self.cloud_cover)}")
                for n in range(self.num_aerosol_modes):
                    file.write("      1")
                file.write("\n")

    def _generate_parah2_ref(self):
        """Generate the parah2.ref file from a profile in the eleos/data/[planet] directory, with any
        pressure limits imposed by NemesisCore.set_pressure_limits()
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            parah2.ref"""
        df = pd.read_table(constants.PATH / f"data/{self.planet}/parah2.ref", skiprows=1, header=None, sep="\s+")
        df.columns = ["height", "parah2"]
        minh, maxh = self.get_height_limits()
        df = df[(df.height >= minh) & (df.height <= maxh)]
        with open(self.directory / "parah2.ref", mode="w+") as file:
            file.write(str(len(df)) + "\n")
            file.write(df.to_string(header=False, index=False))

    def _generate_makephase_inp(self):
        """Add the optical properties of any attached AerosolProfiles to the Makephase input file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            makephase.inp
            normxsc.inp"""
        
        # Get wavelength bounds including the specified reference wavelength
        start, end, step, idx = self._get_xsc_wavelengths(step=0.1)
        
        # Generate the makephase.inp file
        with open(self.directory / "makephase.inp", mode="w+") as file:
            file.write(str(self.num_aerosol_modes) + "\n")
            file.write("1\n")
            utils.write_nums(file, start, end, step)
            file.write("nemesis\n")
            file.write("y\n")

            for name, profile in self.profiles.items():
                if isinstance(profile, profiles_.AerosolProfile):
                    file.write(f"1\n{profile.radius} {profile.variance}\n2\n")
                    if not profile.lookup:
                        file.write(f"1\n{profile.real_n} {profile.imag_n}\n")
                    else:
                        file.write(f"{constants.MAKEPHASE_GASES[profile.n_lookup]}\n")

        # Generate the normxsc.inp file
        with open(self.directory / "normxsc.inp", mode="w+") as file:
            file.write(f"nemesis.xsc\n{idx + 1} 1")

    def _generate_pressure_lay(self):
        """Generate the pressure.lay file that splits the .ref file up into layers for the model
        
        Args:
            min_pressure (float): The pressure at the top of the atmosphere
            max_pressure (float): The pressure at the bottom of the atmosphere
            
        Returns:
            None
            
        Creates:
            pressure.lay"""
        
        pressures = np.logspace(np.log10(self.max_pressure), 
                                np.log10(self.min_pressure), 
                                self.num_layers)
        
        with open(self.directory / "pressure.lay", mode="w+") as file:
            file.write("Created by Eleos\n")
            file.write(str(self.num_layers) + "\n")
            for p in pressures:
                file.write(str(p) + "\n")

    def _generate_summary(self):
        """Dump the information used to create the NEMESIS core to a human-readable text file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            summary.txt
            aerosol_names.txt"""

        with open(self.directory / "summary.txt", mode="w+") as file:
            file.write(f"Generated: {time.asctime()}\n")
            out = ""
            for k, v in self.__dict__.items():
                if k == "profiles":
                    out += "profiles:\n"
                    for label, profile in v.items():
                        out += utils.indent(label, level=1) + "\n"
                        for p in profile.__dict__:
                            if p in ("core", "retrieved"):
                                continue
                            out += utils.indent(f"{p}: {getattr(profile, p)}", level=2) + "\n"

                            if p == "shape":
                                for pp in profile.shape.__dict__:
                                    out += utils.indent(f"{pp}: {getattr(profile.shape, pp)}", level=3) + "\n"
                else:
                    out += f"{k}: {v}"
                out += "\n"
            file.write(out)

        with open(self.directory / "aerosol_names.txt", mode="w+") as file:
            if self.num_aerosol_modes == 0:
                return
            for name, profile in self.profiles.items():
                if isinstance(profile, profiles_.AerosolProfile):
                    file.write(name + "\n")
            file.truncate(file.tell() - 1)
        
    def _get_xsc_wavelengths(self, step):
        """Get the wavelengths for use in Makephase, including the full data range and, if necessary, 
        expanding to include the reference wavelength specified. 
        
        Args:
            step: Wavelength step size (microns)
            
        Returns:
            float: Start wavelength
            float: End wavelength
            float: Wavelength step
            int: reference wavelength index
        """

        # Get wavelengths from spx file
        wls = spx.read(self.spx_file).geometries[0].wavelengths

        # Initial guesses for start and end wavelengths
        start_wl = math.floor(min(wls) / step) * step
        end_wl = math.ceil(max(wls) / step) * step

        # Expand the range to include the reference wavelength if necessary
        if self.reference_wavelength < start_wl:
            start_wl = math.floor(self.reference_wavelength / step) * step  # Round down to nearest step
        elif self.reference_wavelength > end_wl:
            end_wl = math.ceil(self.reference_wavelength / step) * step  # Round up to nearest step
        
        # Find the reference wavelength index and set the reference value to the closest value
        idx, ref = utils.find_nearest(np.arange(start_wl, end_wl+step, step), self.reference_wavelength)
        self.reference_wavelength = ref

        return start_wl, end_wl, step, idx

    def _reset_aerosol_numbering(self):
        """Reset the aerosol numbering. Useful if loading multiple AerosolProfiles from different cores
        
        Args:
            None
            
        Returns:
            None"""
        self.num_aerosol_modes = 0
        for label, profile in self.profiles.items():
            if isinstance(profile, profiles_.AerosolProfile):
                self.num_aerosol_modes += 1
                profile.aerosol_id = self.num_aerosol_modes

    def generate_core(self, verbosity=1, confirm=True):
        """Create all the files necessary for a NEMESIS retrieval in the directory specified by NemesisCore.directory.
        
        Args:
            verbosity (int): One of 0, 1, 2. If 0 then do not print any progress. If 1 then print core number. If 2 then print status for each file
            
        Returns:
            None

        Creates:
            See _generate_* methods
        """
        if verbosity == 1: print(f"Generating core {self.id_}")

        if verbosity == 2: print("Creating directory structure")
        self._create_directory_tree(confirm)

        if verbosity == 2: print(f"Generating summary file")
        self._generate_summary()

        if verbosity == 2: print(f"Copying .spx and .ref files")
        self._copy_input_files()

        if verbosity == 2: print(f"Copying boilerplate files")
        self._copy_template_files()

        if verbosity == 2: print(f"Generating pressure.lay")
        self._generate_pressure_lay()

        if verbosity == 2: print(f"Generating nemesis.sol")
        self._generate_sol()

        if verbosity == 2: print(f"Generating nemesis.cia")
        self._generate_cia()

        if verbosity == 2: print(f"Generating nemesis.inp")
        self._generate_inp()
        
        if verbosity == 2: print(f"Generating nemesis.set")
        self._generate_set()

        if verbosity == 2: print(f"Generating nemesis.fla")
        self._generate_flags()

        if verbosity == 2: print(f"Generating aerosol.ref")
        self._generate_aerosol_ref()

        if verbosity == 2: print(f"Generating parah2.ref")
        self._generate_parah2_ref()

        if verbosity == 2: print(f"Generating nemesis.kls")
        self._generate_kls()

        if verbosity == 2: print(f"Generating fmerror.dat")
        self._generate_fmerror()

        if verbosity == 2: print(f"Generating fcloud.ref")
        self._generate_fcloud_ref()

        if verbosity == 2: print(f"Generating makephase.inp and normxsc.inp")
        self._generate_makephase_inp()

        if verbosity == 2: print(f"Generating nemesis.xsc and [hg]phase.dat files")
        self._generate_xsc()

        if verbosity == 2: print(f"Generating nemesis.apr")
        self._generate_apr()

        if verbosity == 2: print("Saving Eleos core object")
        self._save_core()        

    def run(self):
        """Run NEMESIS on the core. This method should only be used for short forward models as
        it does not schedule the jobs on the ALICE compute nodes (see run_alice_job() for that), 
        instead running it in the current terminal.
        
        Args:
            None
            
        Returns:
            NemeisResult: The results object for the core"""
        
        self.generate_core()
        print("Running NEMESIS...")
        subprocess.run("Nemesis < nemesis.nam > nemesis.prc", 
                       shell=True, 
                       cwd=self.directory,
                       stdout=subprocess.PIPE)
        res = results.NemesisResult(self.directory)
        res.make_summary_plot()
        print("Finished!")
        return res

    def change_parent_directory(self, new_directory):
        raise NotImplementedError("havent got round to this yet")

    def generate_prior_distributions(self):
        """Run NEMESIS briefly in a temporary directory to generate the prior gas and aerosol profiles.
        
        Args:
            None
            
        Returns:
            pd.DataFrame: Data from the .prf files generated. Columns are: 'height', 'pressure', 'temperature', the gas profiles, and the aerosol profiles
        """

        # Change core directory to eleos/tmp temporarily
        old_dir = self.directory
        self.directory = constants.PATH / "tmp"

        # Clear tmp directory
        for fp in self.directory.glob("*"):
            if fp.is_file():
                fp.unlink()

        # Copy any extra files from the main directory
        for filename in old_dir.glob("*"):
            if filename.is_file():
                shutil.copy(filename, self.directory)

        # Generate the core in this directory
        self.generate_core(verbosity=0, confirm=False)

        # Run NEMESIS
        with open(self.directory / "nemesis.nam") as inp:
            proc = subprocess.Popen("Nemesis < nemesis.nam > nemesis.prc", 
                                    shell=True, 
                                    cwd=self.directory,
                                    preexec_fn=os.setsid)
            # Wait for the aerosol.prf file to be generated - potential race condition??
            while True:
                time.sleep(0.05)
                if (self.directory / "aerosol.prf").exists() and (self.directory / "nemesis.prf").exists():
                    break
        
        # Kill NEMESIS
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)

        # Read in the .prf files
        aerosol_prf = parsers.AerosolPrf(self.directory / "aerosol.prf")
        nemesis_prf = parsers.NemesisPrf(self.directory / "nemesis.prf")

        # Reset the directory and clear tmp
        for fp in self.directory.glob("*"):
            if fp.is_file():
                fp.unlink()
        self.directory = old_dir

        nemesis_prf.data.height = np.round(nemesis_prf.data.height * 1000).astype(int) 
        aerosol_prf.data.height = np.round(aerosol_prf.data.height * 1000).astype(int) 
        merged = pd.merge(nemesis_prf.data, aerosol_prf.data, on="height")
        merged.height /= 1000

        return merged

    def get_aerosol_profile(self, id=None):
        """Given an aerosol ID (as used by NEMESIS) return the corresponding AerosolProfile object.
        
        Args:
            id (int): The aerosol ID
            label (str): The label associated with the AerosolProfile object
            
        Returns:
            AerosolProfile: The corresponding aerosol profile"""

        id = abs(id)

        if id > self.num_aerosol_modes:
            raise IndexError(f"Aerosol index is too large! ({id} vs {self.num_aerosol_modes} max.)")

        for profile in self.profiles.values():
            if isinstance(profile, profiles_.AerosolProfile):
                if profile.aerosol_id == id:
                    return profile

    def add_profile(self, profile):
        """Add a profile to retrieve. It's recommended to do this during instantiation using the profiles argument.

        Args:
            profile: profiles.Profile object to add

        Returns:
            None
        """
        self.profiles[profile.label] = profile
        profile.core = self

        if isinstance(profile, profiles_.AerosolProfile):
            # Increment aerosol mode counter        
            self.num_aerosol_modes += 1

            # Assign aerosol IDs
            profile.aerosol_id = self.num_aerosol_modes

    def remove_profile(self, profile_label):
        """Remove a profile. WIP - DOES NOT WORK FOR AEROSOL PROFILES
        
        Args:
            profile_label (str): The label of the profile to remove
            
        Returns:
            None
        """
        profile = self.profiles[profile_label]

        if isinstance(profile, profiles_.AerosolProfile):
            # Decrement aerosol mode counter        
            self.num_aerosol_modes -= 1

        del self.profiles[profile_label]

    def fix_peak(self, central_wavelength, width, error=1e-15):
        """Set a region of the spectra to be fixed (ie. NEMESIS will be forced to match it). This is done internally by setting the
        error on that region to be extrmemely small.
        
        Args:
            central_wavelength (float): The central wavelength of the peak
            width (float): The width of the spectral feature (this is a full width, so the region being fixed is centre-width/2 to centre+width/2)
            error (float): Optionally, give the error to be assigned to the region. This should be very small and there is not much need to change this
        
        Returns:
            None
        """
        self.fixed_peaks.append(FixedPeak(central_wavelength, width, error))

    def exclude_gases(self, *gases):
        """Exclude the given gases by removing the k-tables. This assumes the k-table filepaths are of the form
        **/<gas_name>.combi.kta. Use get_ktable_gas_names() to get a list of all the gases available."""
        self.excluded_gases = gases

    def get_ktable_gas_names(self):
        with open(self.ktable_path) as file:
            lines = file.readlines()
            gas_names = []
            for line in lines:
                match = re.search(r'([^/]+)(?=\.combi\.kta)', line)
                if match:
                    gas_names.append(match.group(1))
        return gas_names

    def set_pressure_limits(self, max_pressure, min_pressure):
        """Change the pressure limits of the core. Also sets the bottom layer height to the closest value in the .ref file
        
        Args:
            max_pressure (float): The new maximum pressure in bar. If None then use the maximum pressure in the .ref file
            min_pressure (float): The new minimum pressure in bar. If None then use the minimum pressure in the .ref file
            
        Returns:
            None"""
        
        if min_pressure is None:
            self.min_pressure = self.ref.data.pressure.min()
        else:
            self.min_pressure = min_pressure
        if max_pressure is None:
            self.max_pressure = self.ref.data.pressure.max()
        else:
            self.max_pressure = max_pressure

        if min_pressure >= max_pressure:
            raise ValueError("min_pressure must be less than max_pressure!")

        i, _ = utils.find_nearest(self.ref.data.pressure, self.max_pressure)
        self.bottom_layer_height = self.ref.data.height.iloc[i]

    def get_height_limits(self):
        """Get the heights of the top and bottom layers of the atmosphere in km.
        
        Args:
            None
            
        Returns:
            top_height: Height of the top of the atmosphere
            bottom_height: Height at the bottom of the atmosphere"""
        
        heights = self.ref.data.height
        return min(heights), max(heights)

    def set_random_priors(self):
        """For each parameter in the associated profiles, set the values randomly based on a normal distribution with mean and standard deviation given by the pre-existing values. 
        
        For example, if there is a single attached GasProfile with the following parameters:
        GasProfile(gas_name="PH3", 
                   shape=shapes.Shape20(
                   knee_pressure=1.0, 
                   tropopause_pressure=0.1,
                   deep_vmr=1.86e-6,          deep_vmr_error=0.2e-6,
                   fsh=0.3,                   fsh_error=0.1))
        then calling this function will set deep_vmr to a value randomly sampled from a normal distribution with mean 1.86e-6 and standard deviation 0.2e-6, and 
        fsh to a value randomly sampled from a normal distribution with mean 0.3 and standard deviation 0.1. The core can then be modified / generated as normal.
        
        Args:
            None

        Returns:
            None
        """
        
        print("Generating random priors...")
        for label, profile in self.profiles.items():
            for name in profile.NAMES:
                try:
                    mean = getattr(profile, name)
                    sigma = getattr(profile, name + "_error")
                    n = -1
                    while n < 0:
                        n = np.random.normal(mean, sigma)
                    if sigma == 0:
                        setattr(profile, name+"_error", 1e-8)
                    setattr(profile, name, n)
                except AttributeError:
                    continue
        print("Generated priors!")

    def create_arbitrary_file(self, filepath, *cols, header=None):
        """Create a file with the pressures in the first columns and then a set of lists for each subsequent column
        
        Args:
            filepath (str): The path to the new file to be created
            *cols List[float]: The values to put in each column
            
        Returns:
            None"""

        with open(filepath, mode="w+") as file:
            if header is not None:
                file.write(header + "\n")
            for i, p in enumerate(self.ref.data.pressure):
                values = [p] + [c[i] for c in cols]
                line = "  ".join([f"{x:.8e}" for x in values])
                print(line)
                file.write(line + "\n")

    def get_profile_variable_names(self):
        """Return the names of the variables (ie the parameters that NEMESIS can vary) in each profile. 
        See also: `get_profile_constant_names` and get_profile_parameter_names`
        
        Args:
            None
            
        Returns:
            dict: Dictionary of the form {profile_label: [variable names, ...]}"""
        
        out = dict()
        for label, profile in self.profiles.items():
            out[label] = profile.VARIABLES
        return out

    def get_profile_constant_names(self):
        """Return the names of the constants (ie. the parameters that NEMESIS doesnt fit) in each profile. 
        See also: `get_profile_constant_names` and get_profile_parameter_names`
        
        Args:
            None
            
        Returns:
            dict: Dictionary of the form {profile_label: [variable names, ...]}"""
        
        out = dict()
        for label, profile in self.profiles.items():
            out[label] = profile.CONSTANTS
        return out
    
    def get_profile_parameter_names(self):
        """Return the names of the parameters (variables and constants) in each profile. 
        See also: `get_profile_constant_names` and get_profile_parameter_names`
        
        Args:
            None
            
        Returns:
            dict: Dictionary of the form {profile_label: [variable names, ...]}"""
        
        out = dict()
        for label, profile in self.profiles.items():
            out[label] = profile.VARIABLES
            out[label] += profile.CONSTANTS
        return out
    

class FixedPeak:
    """Used internally to specify if any spectral regions should be fixed so that NEMESIS always fits it there. Don't instantiate, instead use :meth:`~eleos.cores.NemesisCore.fix_peak`"""
    def __init__(self, central_wavelength, width, error):
        self.central_wavelength = central_wavelength
        self.width = width
        self.error = error

    def __str__(self):
        return f"<FixedPeak at {self.central_wavelength}±{self.width/2:.4f}um>"

    def isin(self, wl):
        """Check whether a specific wavelength is within the region.
        
        Args:
            wl (float): The wavelength to check
        """
        return (wl > self.central_wavelength - self.width/2) & (wl < self.central_wavelength + self.width/2)
    

def add_to_raddata(filepath):
    """Copy a file into the local raddata/ directory. Useful for choosing arbitrary solar spectrum files for example."""
    return shutil.copy(filepath, constants.NEMESIS_PATH / "radtrancode/raddata")


def load_core(core_directory):
    """Load a NemesisCore object saved using NemesisCore._save_core. Do not use this to load a core that has been retrieved;
    use `results.NemesisResult` for that purpose
    
    Args:
        None
        
    Returns:
        NemesisCore: The unpickled core"""
    
    core_directory = Path(core_directory)

    with open(core_directory / "core.pkl", 'rb') as f:
        core = pickle.load(f)

    # Refresh the directory attributes in the NemesisCore object in case the folder has been moved
    core.parent_directory = (core_directory / "..").resolve()
    core.directory = core_directory.resolve()

    return core


def load_from_previous(previous_directory, parent_directory, confirm=True):
    """Load a core from a previous retrieval to use as a template for creating a new core. This method loads the core
    using the core.pkl file and then calls from_previous_retrieval on all Profile objects stored, as well as resetting core ID,
    parent_directory, etc...
    
    Args:
        previous_directory (str): The path to the core directory to load
        parent_directory: THe new parent directory to create the new core under
        
    Returns:
        NemesisCore: The new core object"""
    
    core = load_core(previous_directory)
    res = results.NemesisResult(previous_directory)

    global CORE_ID_COUNTER
    CORE_ID_COUNTER += 1
    core.id_ = CORE_ID_COUNTER

    core.parent_directory = Path(parent_directory)
    core.directory = core.parent_directory / f"core_{core.id_}"

    core.num_aerosol_modes = 0
    for label, profile in core.profiles.items():
        core.add_profile(profile.from_previous_retrieval(res, label))

    return core


def create_sensitivity_analysis(parent_directory, 
                                template_core,
                                generate_cores=True, 
                                factors=(0.80, 0.90, 0.95, 1.05, 1.10, 1.20), 
                                forward=True):
    
    """
    Create a sensitivity analysis on the template core. This is done by creating a new core for 
    each parameter in each profile, and varying that parameter by a small amount.
    
    Args:
        parent_directory (str):      The parent directory to create the new cores in
        template_core (NemesisCore): The core to use as a template
        generate_cores (bool):       Whether to generate the cores or just return the list of cores to be generated
        factors (tuple):             The factors to vary the parameters by. Default is (0.8, 0.9, 0.95, 1.05, 1.1, 1.2). Note there is no 
                                     need to include 1.00 as a factor as this is equivalent to the baseline which is already calculated
        forward (bool):              Whether to run forward models or retrievals. Use forward models to see where each parameter affects each 
                                     spectrum and retrievals to test if different priors converge on different solutions
    
    Returns:
        List[NemesisCore]: A list of new cores with the parameters varied
    """
    
    global CORE_ID_COUNTER
    reset_core_numbering()
    parent_directory = Path(parent_directory)
    parent_directory.mkdir(parents=True)

    # Create a file that stored which parameters were varied and by how much for each core
    file = open(f"{parent_directory}/sensitivity_analysis.txt", "w+")
    file.write("Core ID,Profile Label,Parameter,Base Value,Shifted Value,Factor\n")

    CORE_ID_COUNTER += 1
    template_core.parent_directory = parent_directory
    template_core.directory = parent_directory / f"core_{CORE_ID_COUNTER}"
    template_core.forward = forward
    cores = [template_core]

    # Generate the template core if requested
    if generate_cores:
        template_core.forward = True
        template_core.generate_core()

    # Iterate through every parameter in every profile in the template core
    for label, params in template_core.get_profile_parameter_names().items():
        file.flush()

        for param in params:
            # Vary the parameter value by 80% to 120%
            for factor in factors:

                # Copy the core (deep copy to make sure nothing is passed by reference)
                core = copy.deepcopy(template_core)

                # Increment core ID counter manually
                CORE_ID_COUNTER += 1
                core.id_ = CORE_ID_COUNTER

                # Set to forward mode and change the core directory
                core.parent_directory = parent_directory
                core.directory = parent_directory / f"core_{core.id_}"
                core.forward = forward

                # Get and modify the parameter value and log it in the parameter file
                base_value = getattr(core.profiles[label], param)
                setattr(core.profiles[label], param, base_value * factor)
                cores.append(core)
                file.write(f"{core.id_},{label},{param},{base_value},{base_value * factor},{factor}\n")

                # Generate the cores if requested
                if generate_cores:
                    core.generate_core()

    # Always close your files, kids!
    file.close()
    
    return cores


def create_gas_analysis_cores(parent_directory, template_core, generate_cores=True, new_spx=None):
    """Create a set of cores where each core has a single gas excluded. USeful for determining where in 
    the spectrum each gas contributes.
    
    Args:
        parent_directory (str): The parent directory to create the new cores in
        template_core (NemesisCore): The core to use as a template
        generate_cores (bool): Whether to generate the cores or just return the list of cores to be generated
        new_spx (str): The path to a new spx file to use instead of the one in the template core. This is useful if you want to use a different resolution or wavelength range

    Returns:
        List[NemesisCore]: A list of new cores with the gases excluded"""
    
    def make_core(template_core, parent_directory, new_spx):
        global CORE_ID_COUNTER
        core = copy.deepcopy(template_core)
        CORE_ID_COUNTER += 1
        core.id_ = CORE_ID_COUNTER

        core.parent_directory = Path(parent_directory)
        core.directory = core.parent_directory / f"core_{core.id_}"
        core.forward = True

        if new_spx is not None:
            core.spx_file = Path(new_spx)

        return core

    clear_parent_directory(parent_directory)
    reset_core_numbering()

    all_cores = [make_core(template_core, parent_directory, new_spx)]

    names = template_core.get_ktable_gas_names()
    for name in names:
        core = make_core(template_core, parent_directory, new_spx)
        core.exclude_gases(name)
        all_cores.append(core)
    
    if generate_cores:
        for core in all_cores:
            core.generate_core()
        
    return all_cores


def reset_core_numbering(): 
    """Reset the automatic core numbering. Useful for creating multiple sets of cores in one program.
    
    Args:
        None
        
    Returns:
        None"""
    global CORE_ID_COUNTER
    CORE_ID_COUNTER = 0


def reset_core(core_directory):
    """Delete all the files that NEMESIS generates and leave only the input files. 
    Useful for re-running a retrieval from scratch."""

    core_directory = Path(core_directory)
    to_delete = [
        "kk.dat",
        "kk.out",
        "nemesis.chi",
        "nemesis.cov",
        "nemesis.drv",
        "nemesis.itr",
        "nemesis.log",
        "nemesis.mre",
        "nemesis.pat",
        "nemesis.prc",
        "nemesis.prf",
        "nemesis.raw",
        "nemesis.sca",
        "aerosol.prf",
        "fcloud.prf",
        "*slurm*",
        "parah2.prf",
        "refindex4.dat",
        "refindex6.dat"
    ]
    for pattern in to_delete:
        for file_path in core_directory.glob(pattern):
            if file_path.is_file():
                file_path.unlink()
                print(f"Deleted: {file_path.name}")


def clear_parent_directory(parent_directory, confirm=True):
    """Attempt to delete all the child files/directories from the specified folder. It is recommended to call this function at the start
    of every core generation script as it ensures that the script is fully stateless. If a script generates N cores when first run
    and is subsequently modified to produce N/2 cores, then the remaining N/2 core directories will remain in the parent directory.
    This should not matter as the SLURM script will only run the correct subset of cores, but it can get confusing. If the parent directory
    specified does not exist then it will print a warning and return.
    
    Args:
        parent_directory (str): The directory to clear
        confirm (bool): Whether to confirm with the user before deleting the directory
        
    Returns:
        None"""
    
    parent_directory = Path(parent_directory)
    if os.path.exists(parent_directory):
        if confirm:
            x = input(f"{parent_directory.resolve()} already exists - are you sure you want to erase and continue? Y/N ")
            if x.lower() == "n":
                print("Quitting")
                exit()
            else:
                shutil.rmtree(parent_directory)
        else:
            shutil.rmtree(parent_directory)


def get_refractive_indicies(name, start_wl, end_wl, wl_step):
    """Run Makephase to get the refractive indicies of the gases in the lookup tables.
    
    Args:
        name (str): The gas name to use (see constants.MAKEPHGASE_GASES for a list)
        start_wl: Start wavelength
        end_wl: End wavelength
        wl_step: Wavelength step size
        
    Returns:
        pd.DataFrame: Dataframe containing the wavelength and the real and imaginary refractive indicies. Columns are: 'wavelength', 'real', and 'imag'
    """
    
    # Make Makephase input file
    with open(constants.PATH / "tmp/makephase.inp", mode="w+") as file:
        file.write("1\n1\n")
        utils.write_nums(file, start_wl, end_wl, wl_step)
        file.write("tmp.xsc\ny\n1\n1\n1\n2\n")
        file.write(str(constants.MAKEPHASE_GASES[name]))
        if name == "Tholins":
            file.write("\n1")

    # Run Makephase
    proc = subprocess.run(f"Makephase < makephase.inp", 
                           shell=True,
                           cwd=constants.PATH.resolve() / "tmp",
                           capture_output=True)
    
    # Extract refractive indicies from stdout
    real, imag = [], []
    for line in proc.stdout.decode("ASCII").split("\n"):
        if line.startswith(" Refractive index:"):
            r, i = utils.get_floats_from_string(line)
            real.append(r)
            imag.append(i)

    # Clean up tmp/
    for fp in (constants.PATH / "tmp/").glob("*"):
        if fp.is_file():
            fp.unlink()

    wavelengths = np.arange(start_wl, end_wl+wl_step, wl_step)
    data = pd.DataFrame(np.array([wavelengths, real, imag]).T, columns=["wavelength", "real", "imag"])
    return data


def generate_alice_job(parent_directory, 
                       python_env_name, 
                       memory=16, 
                       hours=24,
                       username=None,
                       notify=("end", "fail"),
                       type_="normal"):
    """Generate an sbatch submission script for use on ALICE. The job is an array job over all the specified cores.
    After running NEMESIS it will run Eleos to create some summary plots in the parent/core_N/plots directory
    
    Args:
        parent_directory (str): The directory containing the NEMESIS core directories to be run
        python_env_name (str): The name of a conda environment which has Eleos installed
        memory (int): The amount of memory to use (in GB)
        hours (int): The number of hours to schedule the job for
        username (str): The username of the user running the job (eg, scat2, lnf2)
        notify (List[str] or str): What email notifications to send. Can be any combination of 'begin', 'end', 'fail'
        type_ (str): The type of job to run. Can be 'normal' or 'sensitivity'. This will determine how Eleos is run at the end of the job.
        
    Returns:
        None
        
    Creates:
        submitjob.sh in the parent directory of the cores"""

    parent_directory = Path(parent_directory)
    script_path = parent_directory / "submitjob.sh"
    ncores = len([f for f in os.listdir(parent_directory) if not os.path.isfile(parent_directory / f)])

    # Coerce notify to an iterable and create the string
    try:
        len(notify)
    except:
        notify = [notify]
    notify_str = ",".join(notify).upper()   

    # Assert that hours and memory are ints
    assert isinstance(hours, int)
    assert isinstance(memory, int) 

    # Read the submission script and replace template fields
    with open(constants.PATH / "data/statics/template.job", mode="r") as file:
        out = file.read()
        out = out.replace("<MEMORY>", str(memory))
        out = out.replace("<HOURS>", f"{hours:02}")
        out = out.replace("<N_CORES>", str(ncores))
        out = out.replace("<CORE_DIR>", str(parent_directory.resolve()))
        out = out.replace("<USERNAME>", str(username))
        out = out.replace("<PYTHON_ENV>", python_env_name)
        out = out.replace("<NOTIFY>", notify_str)

    if type_ == "sensitivity":
        out += f"\npython -m eleos {parent_directory.resolve()} --make-sensitivity-summary --run-if-finished"

    # Write the filled template 
    with open(script_path, mode="w+") as file:
        file.write(out)


def run_alice_job(parent_directory, print_queue_delay=2):
    """Run the script created by generate_alice_job
    
    Args:
        parent_directory (str): Path to the directory containing all the cores
        print_queue_delay (float): Duration to wait (in seconds) before printing the current job queue

    Returns:
        None
    """
    parent_directory = Path(parent_directory)
    os.system(f"sbatch {parent_directory / 'submitjob.sh'}")
    time.sleep(print_queue_delay)
    os.system("squeue --me")



import argparse
import subprocess
from pathlib import Path

from . import results


def exit_msg(msg="Not all cores have finished running!"):
    print(msg)
    exit()


def check_all_finished(directory):
    for core in directory.glob("**/nemesis.apr"):
        try:
            slurm = tuple((core / "..").resolve().glob("*slurm*.out"))[0]
            if "Done!" not in open(slurm).read():
                exit_msg()
        except IndexError:
            exit_msg()
    print("All cores have run! Running command...")


def make_summary(core_dir, log=False, silent=False):
    res = results.NemesisResult(core_dir)
    if not silent:
        res.print_summary()
    res.make_summary_plot(log=log)
    if not res.core.forward:
        res.make_iterations_plot()
    print("Done!")


def make_sensitivity_summary(parent_dir):
    sens = results.SensitivityAnalysis(parent_dir)
    sens.make_parameters_plot()
    sens.savefig("sensitivity.png", dpi=400)
    print("Done!")


parser = argparse.ArgumentParser(
    description="Run eleos analysis commands."
)
subparsers = parser.add_subparsers(dest="command", required=True)

# make-summary command
summary_parser = subparsers.add_parser("make-summary", help="Print summary tables and generate plots for a core")
summary_parser.add_argument("directory",         help="Path to a core directory",)
summary_parser.add_argument("--run-if-finished", help="Run only if all cores in directory are finished",       action="store_true",)
summary_parser.add_argument("--log-spectrum",    help="Whether to plot the spectrum/residuals on a log scale", action="store_true",)
summary_parser.add_argument("--silent",          help="Whether to print a table of the priors/fitted parameters", action="store_true",)

# make-sensitivity-summary command
sens_parser = subparsers.add_parser("make-sensitivity-summary", help="Create a plot showing effect of changing parameters on the spectrum")
sens_parser.add_argument("directory",         help="Path to a parent directory containing sensitivity analysis",)
sens_parser.add_argument("--run-if-finished", help="Run only if all cores in directory are finished", action="store_true")


args = parser.parse_args()
directory = Path(args.directory)

if args.run_if_finished:
    check_all_finished(directory)

if args.command == "make-summary":
    make_summary(directory, log=args.log_spectrum, silent=args.silent)

elif args.command == "make-sensitivity-summary":
    make_sensitivity_summary(directory)

else:
    parser.error("Command not recognised!")

